from .add import scatter_add
from .sub import scatter_sub
from .mul import scatter_mul
from .div import scatter_div
from .mean import scatter_mean
from .std import scatter_std
from .max import scatter_max
from .min import scatter_min

__version__ = '1.1.2'

__all__ = [
    'scatter_add',
    'scatter_sub',
    'scatter_mul',
    'scatter_div',
    'scatter_mean',
    'scatter_std',
    'scatter_max',
    'scatter_min',
    '__version__',
]
